/**
 *#########################################################################
 *
 * A component of the Gatherer application, part of the Greenstone digital
 * library suite from the New Zealand Digital Library Project at the
 * University of Waikato, New Zealand.
 *
 * <BR><BR>
 *
 * Author: John Thompson, Greenstone Digital Library, University of Waikato
 *
 * <BR><BR>
 *
 * Copyright (C) 1999 New Zealand Digital Library Project
 *
 * <BR><BR>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * <BR><BR>
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * <BR><BR>
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *########################################################################
 */

 




/* GPL_HEADER */
package org.greenstone.gatherer.cdm;
/**************************************************************************************
 * Title:        Gatherer
 * Description:  The Gatherer: a tool for gathering and enriching a digital collection.
 * Company:      The University of Waikato
 * Written:      01/05/02
 * Revised:      16/08/02 Optimized and Commented.
 **************************************************************************************/
import java.io.Serializable;
import java.util.ArrayList;
import org.greenstone.gatherer.cdm.Argument;
import org.greenstone.gatherer.cdm.ArgumentContainer;
import org.greenstone.gatherer.cdm.CustomClassifier;
import org.greenstone.gatherer.cdm.Format;
/** This class is responsible for storing information from a parsed classinfo.pl call in such a way that it allows easy access to parsed details for the purposes of user design and specification of classifiers.
 * @author John Thompson, Greenstone Digital Library, University of Waikato
 * @version 2.3
 */
// ####################################################################################
// Optimization                          Saving
// ####################################################################################
// Vector -> ArrayList (x8)              + Processor
// ####################################################################################
public class Classifier
    extends ArrayList
    implements ArgumentContainer, Comparable, Serializable {
    /** A reference to the classifier that this one inherits from. */
    private Classifier super_classifier = null;
    /** A reference to the manager of this classifier. Only available once the classifiers been assigned, always <i>null</i> for classifiers in reserve. */
    private ClassifierManager manager = null;
    /** Custom arguments (ie those unparsable from classinfo.pl) provided to this class. */
    private String custom = null;
    /** A description of this classifier. */
    private String desc = null;
    /** The name of the classifier as it would appear in the collect.cfg file. */
    private String name = null;
    /** A list of format commands that are dependant on this classifier. */
    private ArrayList dependant_formats = null;
    /** Default Constructor.
     */
    public Classifier() {
	super();
	dependant_formats = new ArrayList();
    }
    /** Constructor.
	  * @param name The name of this classifier as a <strong>String</strong>.
	  * @param desc A description of this classifier as a <strong>String</strong>.
	  * @param super_classifier The super class of this classifier, as a <strong>Classifier</strong>.
	  */
    public Classifier(String name, String desc, Classifier super_classifier) {
	this();
	this.desc = desc;
	this.name = name;
	this.super_classifier = super_classifier;
    }
    /** Method to add an argument to this classifier. Only adds the argument if it isn't already present.
	  * @param argument The <strong>Argument</strong> to add.
	  */
    public void addArgument(Argument argument) {
	if(!contains(argument)) {
	    add(argument);
	    argument.setOwner(name);
	}
    }
    /** Method to register a dependant format with this classifier. If the classifier changes, this format also needs to be updated.
	  * @param format A <strong>Format</strong> which is dependant on this classifier.
	  */
    public void addDependantFormat(Format format) {
	dependant_formats.add(format);
    }
    /** Method to compare two classifiers for ordering.
	  * @param object The classifier we are comparing to, as an <strong>Object</strong>.
	  * @return An <i>int</i> specifying the classifier order, using values as set out in String.
	  * @see java.lang.String#compareTo
	  */
    public int compareTo(Object object) {
	if(object instanceof Classifier) {
	    Classifier classifier = (Classifier) object;
	    return name.compareTo(classifier.getName());
	}
	return name.compareTo(object.toString());
    }
    /** This method produces a deep copy of this classifier. Note that this also creates a new copy of each of the super classes of classifiers as well. This is the way it should be, as each assigned classifier may have different values for the higher classifiers (such as BasPlug).
	  * @return A newly created <strong>Classifier</strong> with the same details and Arguments as this one.
	  * @see org.greenstone.gatherer.cdm.Argument
	  */
    public Classifier copy() {
	Classifier copy = null;
	if(super_classifier == null) {
	    copy = new Classifier(name, desc, null);
	}
	else {
	    copy = new Classifier(name, desc, super_classifier.copy());
	}
	for(int i = 0; i < size(); i++) {
	    copy.addArgument(((Argument)get(i)).copy());
	}
	return copy;
    }
    /** Method to determine if two classifiers are equal.
	  * @param object The classifier to test against, as an <strong>Object</strong>.
	  * @return <i>true</i> if the classifier names match, <i>false</i> otherwise.
	  * @see org.greenstone.gatherer.cdm.CustomClassifier
	  */
    public boolean equals(Object object) {
	if(object instanceof CustomClassifier) {
	    CustomClassifier classifier = (CustomClassifier) object;
	    return (toString().equalsIgnoreCase(classifier.getCommand()));
	}
	else {
	    return(toString().equalsIgnoreCase(object.toString()));
	}
    }
    /** Method to retrieve an argument by its name.
	  * @param name The name of the argument as a <strong>String</strong>.
	  * @return The <strong>Argument</strong> requested, or <i>null</i> if no such argument.
	  */
    public Argument getArgument(String name) {
	// The name given may still include the '-'
	if(name.startsWith("-")) {
	    name = name.substring(1);
	}
	ArrayList arguments = getArguments();
	for(int i = 0; i < arguments.size(); i++) {
	    Argument argument = (Argument)arguments.get(i);
	    if(argument.getName().equals(name)) {
		return argument;
	    }
	}
	return null;
    }
    /** Method to retrieve all of the arguments available to a classifier, including both specific and general ones.
	  * @return A <strong>Hashtable</strong> of arguments, with &lt;name&gt; -&gt; &lt;argument&gt; entries.
	  */
    public ArrayList getArguments() {
	ArrayList all_arguments = new ArrayList(this);
	if(super_classifier != null) {
	    ArrayList super_arguments = super_classifier.getArguments();
	    for(int i = 0; i < super_arguments.size(); i++) {
		Object argument = super_arguments.get(i);
		if(!all_arguments.contains(argument)) {
		    all_arguments.add(argument);
		}
	    }
	}
	return all_arguments;
    }
    /** Method to retrieve a classifiers custom argument information.
	  * @return The custom arguments as a <strong>String</strong>.
	  */
    public String getCustom() {
	return custom;
    }
    /** Method to retrieve a classifiers name.
	  * @return A <strong>String</strong> containing the classifiers name.
	  */
    public String getName() {
	return name;
    }
    /** Method to retrieve the position of this classifier as a special keyword or the form "CL#" where # is the classifiers order in the set of classifiers. Note that if this is called for a Classifier that has never been assigned a position of 'Search' is returned, as this is the only case where we will try to write a Classifier that does not have a manager.
	  * @return A <strong>String</strong> containing the special position keyword.
	  * @see org.greenstone.gatherer.cdm.ClassifierManager
	  */
    public String getPositionString() {
	if(manager == null) {
	    return "Search";
	}
	return "CL" + (manager.indexOf(this) + 1);
    }
    /** Method to set the value of custom.
	  * @param custom The new value of custom as a <strong>String</strong>.
	  */
    public void setCustom(String custom) {
	this.custom = custom;
    }
    /** Method to set the value of desc.
	  * @param desc The new value of desc as a <strong>String</strong>.
	  */
    public void setDesc(String desc) {
	this.desc = desc;
    }
    /** Method to set the value of manager.
	  * @param manager The new manager as a <strong>ClassifierManager</strong>.
	  */
    public void setManager(ClassifierManager manager) {
	this.manager = manager;
    }
    /** Method to set the value of name.
	  * @param name The new value of name as a <strong>String</strong>.
	  */
    public void setName(String name) {
	this.name = name;
    }
    /** Method to set the value of the super_classifier.
	  * @param super_classifier The new value of super_classifier as a <strong>Classifier</strong>, or <i>null</i> if this class has no inheritance.
	  */
    public void setSuper(Classifier super_classifier) {
	this.super_classifier = super_classifier;
    }
    /** Method to print out this classifier as it would appear to the user in the interface
	  * @return A <strong>String</strong> containing a single classifier command.
	  */
    public String toString() {
	StringBuffer text = new StringBuffer("classify ");
	text.append(name);
	text.append(" ");
	ArrayList arguments = getArguments();
	for(int i = 0; i < arguments.size(); i++) {
	    Argument argument = (Argument)arguments.get(i);
	    if(argument.isAssigned()) {
		text.append(argument.toString());
		text.append(" ");
	    }
	}
	if(custom != null) {
	    text.append(custom);
	}
	return text.toString();
    }
    /** Method to print out this classifier as it would appear as a command within the collection configuration file.
	  * @return A <strong>String</strong> containing a single classifier command.
	  */
    public String toStringConfig() {
	StringBuffer text = new StringBuffer("classify ");
	text.append(name);
	text.append(" ");
	ArrayList arguments = getArguments();
	for(int i = 0; i < arguments.size(); i++) {
	    Argument argument = (Argument)arguments.get(i);
	    if(argument.isAssigned()) {
		text.append(argument.toStringConfig());
		text.append(" ");
	    }
	}
	if(custom != null) {
	    text.append(custom);
	}
	return text.toString();
    }
}


