/**
 *#########################################################################
 *
 * A component of the Gatherer application, part of the Greenstone digital
 * library suite from the New Zealand Digital Library Project at the
 * University of Waikato, New Zealand.
 *
 * Author: John Thompson, Greenstone Digital Library, University of Waikato
 *
 * Copyright (C) 1999 New Zealand Digital Library Project
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *########################################################################
 */
package org.greenstone.gatherer.cdm;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.table.*;
import org.greenstone.gatherer.Dictionary;
import org.greenstone.gatherer.Gatherer;
import org.greenstone.gatherer.cdm.CollectionDesignManager;
import org.greenstone.gatherer.cdm.CollectionMeta;
import org.greenstone.gatherer.cdm.CollectionMetaManager;
import org.greenstone.gatherer.cdm.Language;
import org.greenstone.gatherer.cdm.LanguageManager;

/** This class provides a graphical interface to allow a user to quickly and conviently (ie all in one place) translate the text fragments associated with general metadata and indexes into each of the assigned languages in the collection. It should provide clear controls for the editing of these text fragments, plus clear indicate what languages still need further translation, which it will do through a combination of coloring and other visual indicators.
 * @author John Thompson, Greenstone Digital Library, University of Waikato
 * @version 2.3
 */
public class TranslationManager {

    private CollectionDesignManager manager;
    private Control controls;

    static final private Dimension LABEL_SIZE = new Dimension(225,25);
    static final private int LANGUAGE_WIDTH = 75;
    static final private String GENERAL_STR = "General:";
    static final private String INDEX_STR = "Index:";
    static final private String PARTITION_STR = "Partitions:";

    TranslationManager(CollectionDesignManager manager) {
	this.manager = manager;
    }

    public JPanel getControls() {
	if(controls == null) {
	    controls = new Control();
	}
	return controls;
    }

    private String get(String key) {
	if(key.indexOf(".") == -1) {
	    key = "CDM.TranslationManager." + key;
	}
	return Gatherer.dictionary.get(key);
    }

    private Object[] getFeaturesList() {
	ArrayList features_model = new ArrayList();
	for(int i = 0; i < manager.collectionmetadatum.size(); i++) {
	    CollectionMeta metadata = (CollectionMeta) manager.collectionmetadatum.get(i);
	    Object bob = new BobTheMagicalComparableWrapper(metadata.getName());
	    if(!features_model.contains(bob)) {
		features_model.add(bob);
	    }
	}
	// We also add subindexes.
	SubIndexes subindexes = manager.subcollections.getSubIndexes();
	for(int j = 0; j < subindexes.size(); j++) {
	    Object bob = new BobTheMagicalComparableWrapper(subindexes.get(j));
	    if(!features_model.contains(bob)) {
		features_model.add(bob);
	    }
	}
	Collections.sort(features_model);
	return features_model.toArray();
    }


    private class BobTheMagicalComparableWrapper
	implements Comparable {
	private Object content;
	private String text;
	BobTheMagicalComparableWrapper(Object content) {
	    this.content = content;
	}
	public int compareTo(Object object) {
	    if(object == null) {
		object = "";
	    }
	    if(text == null) {
		toString();
	    }
	    return text.compareTo(object.toString());
	}
	public boolean equals(Object object) {
	    if(object == null) {
		return false;
	    }
	    if(text == null) {
		toString();
	    }
	    return text.equals(object.toString());
	}
	public Object getContent() {
	    return content;
	}
	public String toString() {
	    if(text == null) {
		if(content instanceof Index) {
		    text = INDEX_STR + content.toString();
		}
		else if(content instanceof SubIndex) {
		    text = PARTITION_STR + content.toString();
		}
		else {
		    text = GENERAL_STR + content.toString();
		}
	    }
	    return text;
	}
    }

    private class Control
	extends JPanel {

	private boolean ignore_event = false;
	private FragmentTableModel fragment_table_model;
	private JComboBox language_combobox;
	private JList features_list;
	private JTable fragment_table;
	private JTextArea default_area;
	private JTextArea translation_area;

	Control() {
	    fragment_table_model = new FragmentTableModel(new ArrayList());
	    // Creation
	    JPanel header_panel = new JPanel();
	    JLabel title_label = new JLabel(get("Title"));
	    title_label.setHorizontalAlignment(JLabel.CENTER);
	    JTextArea instruction_area = new JTextArea(get("Instructions"));
	    instruction_area.setEditable(false);
	    instruction_area.setLineWrap(true);
	    instruction_area.setRows(3);
	    instruction_area.setWrapStyleWord(true);

	    JPanel center_panel = new JPanel();

	    JPanel selection_panel = new JPanel();

	    JPanel component_selection_panel = new JPanel();
	    JLabel component_label = new JLabel(get("Affected_Features"));
	    features_list = new JList(getFeaturesList());
	    features_list.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

	    JPanel fragment_selection_panel = new JPanel();
	    JLabel fragment_label = new JLabel(get("Assigned_Fragments"));

	    fragment_table = new JTable(fragment_table_model);
	    fragment_table.setAutoResizeMode(JTable.AUTO_RESIZE_LAST_COLUMN);
	    fragment_table.setBackground(Gatherer.config.getColor("coloring.collection_tree_background", false));
	    fragment_table.setColumnSelectionAllowed(false);
	    fragment_table.setDefaultRenderer(Object.class, new FragmentTableRenderer());
	    fragment_table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
	    //TableColumnModel column_model = fragment_table.getColumnModel();
	    //TableColumn language_column = column_model.getColumn(0);
	    //language_column.setPreferredWidth(LANGUAGE_WIDTH);
	    //language_column = null;
	    //column_model = null;
	    //fragment_table.sizeColumnsToFit(0);

	    JScrollPane table_scroll = new JScrollPane(fragment_table);
		table_scroll.getViewport().setBackground(Gatherer.config.getColor("coloring.collection_tree_background", false));
		table_scroll.setOpaque(true);

	    JPanel south_panel = new JPanel();

	    JPanel text_panel = new JPanel();

	    JPanel language_panel = new JPanel();
	    JLabel language_label = new JLabel(get("Language"));
	    language_label.setPreferredSize(LABEL_SIZE);
	    language_combobox = new JComboBox(manager.languages.getLanguageCodes().toArray());

	    JPanel default_text_panel = new JPanel();
	    JLabel default_label = new JLabel(get("Default_Text"));
	    default_area = new JTextArea();
	    default_area.setBackground(Gatherer.config.getColor("coloring.collection_tree_background", false));
	    default_area.setEditable(false);
	    default_area.setLineWrap(true);
	    default_area.setWrapStyleWord(true);

	    JPanel translated_text_panel = new JPanel();
	    JLabel translation_label = new JLabel(get("Translation"));
	    translation_area = new JTextArea();
	    translation_area.setBackground(Gatherer.config.getColor("coloring.disabled", false));
	    translation_area.setEnabled(false);
	    translation_area.setLineWrap(true);
	    translation_area.setWrapStyleWord(true);

	    // Connection
	    language_combobox.addActionListener(new LanguageActionListener());
	    translation_area.getDocument().addDocumentListener(new TranslationDocumentListener());
	    features_list.addListSelectionListener(new FeaturesListSelectionListener());
	    fragment_table.getSelectionModel().addListSelectionListener(new FragmentListSelectionListener());
	    // Layout
	    header_panel.setBorder(BorderFactory.createEmptyBorder(0,0,5,0));
	    header_panel.setLayout(new BorderLayout());
	    header_panel.add(title_label, BorderLayout.NORTH);
	    header_panel.add(new JScrollPane(instruction_area), BorderLayout.CENTER);

	    component_selection_panel.setLayout(new BorderLayout());
	    component_selection_panel.add(component_label, BorderLayout.NORTH);
	    component_selection_panel.add(new JScrollPane(features_list), BorderLayout.CENTER);

	    fragment_selection_panel.setLayout(new BorderLayout());
	    fragment_selection_panel.add(fragment_label, BorderLayout.NORTH);
	    fragment_selection_panel.add(table_scroll, BorderLayout.CENTER);

	    selection_panel.setLayout(new GridLayout(2,1,0,5));
	    selection_panel.add(component_selection_panel);
	    selection_panel.add(fragment_selection_panel);

	    default_text_panel.setLayout(new BorderLayout());
	    default_text_panel.add(default_label, BorderLayout.NORTH);
	    default_text_panel.add(new JScrollPane(default_area), BorderLayout.CENTER);

	    translated_text_panel.setLayout(new BorderLayout());
	    translated_text_panel.add(translation_label, BorderLayout.NORTH);
	    translated_text_panel.add(new JScrollPane(translation_area), BorderLayout.CENTER);

	    text_panel.setLayout(new GridLayout(1,2,5,0));
	    text_panel.add(default_text_panel);
	    text_panel.add(translated_text_panel);

	    language_panel.setLayout(new BorderLayout());
	    language_panel.add(language_label, BorderLayout.WEST);
	    language_panel.add(language_combobox, BorderLayout.CENTER);

	    south_panel.setLayout(new BorderLayout());
	    south_panel.add(language_panel, BorderLayout.NORTH);
	    south_panel.add(text_panel, BorderLayout.CENTER);

	    center_panel.setLayout(new GridLayout(2,1,0,5));
	    center_panel.add(selection_panel);
	    center_panel.add(south_panel);

	    setBorder(BorderFactory.createEmptyBorder(5,5,5,5));
	    setLayout(new BorderLayout());
	    add(header_panel, BorderLayout.NORTH);
	    add(center_panel, BorderLayout.CENTER);
	}

	public void updateUI() {
	    // Rebuild the features model, just incase features have been added or removed.
	    if(features_list != null) {
		Object selected_feature = features_list.getSelectedValue();
		features_list.setListData(getFeaturesList());
		if(selected_feature != null) {
		    features_list.setSelectedValue(selected_feature, true);
		}
	    }
	    // Carry on.
	    super.updateUI();
	}

	private class FeaturesListSelectionListener
	    implements ListSelectionListener {
	    public void valueChanged(ListSelectionEvent event) {
		if(!event.getValueIsAdjusting()) {
		    if(!ignore_event) { // And it shouldn't ever be.
			ignore_event = true;
			///ystem.err.println("FeaturesListSelectionListener");
			if(!features_list.isSelectionEmpty()) {
			    // Determine the features name. Remember to remove our helpful source prefix (delimited by ':').
			    BobTheMagicalComparableWrapper bob = (BobTheMagicalComparableWrapper)features_list.getSelectedValue();
			    Object selected_feature = bob.getContent();
			    bob = null;
			    // Retrieve the default language translation, or otherwise the first match, to be default text.
			    CollectionMeta default_metadata = manager.collectionmetadatum.getMetadata(selected_feature, Gatherer.config.interface_language, true);
			    if(default_metadata != null) {
				default_area.setText(default_metadata.getValue());
			    }
			    // Update the text fragment table.
			    fragment_table.clearSelection();
			    fragment_table_model.setData(manager.collectionmetadatum.getMetadata(selected_feature));
			    selected_feature = null;
			    // Now we check whatever value is currently selected in the combobox to see if it is valid to add.
			    String language_name = language_combobox.getSelectedItem().toString();
			    int index = fragment_table_model.getMetadataIndexByLanguage(language_name);
			    if(index != -1) {
				CollectionMeta metadata = fragment_table_model.getMetadata(index);
				fragment_table.setRowSelectionInterval(index, index);
				translation_area.setText(metadata.getValue());
				metadata = null;
			    }
			    else {
				translation_area.setText("");
			    }
			    // Update and enable the text area
			    translation_area.setEnabled(true);
			    translation_area.setBackground(Gatherer.config.getColor("coloring.editable", false));
			}
			else {
			    default_area.setText("");
			    fragment_table.clearSelection();
			    fragment_table_model.setData(new ArrayList());
			    translation_area.setText("");
			    // Update and disable the text area
			    translation_area.setText("");
			    translation_area.setEnabled(false);
			    translation_area.setBackground(Gatherer.config.getColor("coloring.disabled", false));
			}
			ignore_event = false;
		    }
		}
	    }
	}

	private class FragmentListSelectionListener
	    implements ListSelectionListener {
	    public void valueChanged(ListSelectionEvent event) {
		if(!event.getValueIsAdjusting() && !ignore_event) {
		    ignore_event = true;
		    ///ystem.err.println("FragmentListSelectionListener");
		    int index = -1;
		    if((index = fragment_table.getSelectedRow()) != -1) {
			// A row has been selected. Retrieve the collection metadata.
			CollectionMeta metadata = fragment_table_model.getMetadata(index);
			if(metadata != null) {
			    // Update the combobox to show the current language.
			    Language language = metadata.getLanguage();
			    String language_name = language.toString();
			    ///ystem.err.println("Searching for the language: " + language_name);
			    for(int i = 0; i < language_combobox.getItemCount(); i++) {
				if(language_combobox.getItemAt(i).toString().equals(language_name)) {
				    language_combobox.setSelectedIndex(i);
				}
			    }
			}
			translation_area.setText(metadata.getValue());
		    }
		    else {
			translation_area.setText("");
		    }
		    ignore_event = false;
		}
	    }
	}

	private class FragmentTableModel
	    extends AbstractTableModel {

	    private ArrayList metadatum;

	    FragmentTableModel(ArrayList metadatum) {
		this.metadatum = metadatum;
	    }
	    public int getRowCount() {
		return metadatum.size();
	    }
	    public int getColumnCount() {
		return 2;
	    }
	    public String getColumnName(int column) {
		return get("Column_Name_" + column);
	    }

	    public CollectionMeta getMetadata(int row) {
		if(0 <= row && row < metadatum.size()) {
		    return (CollectionMeta) metadatum.get(row);
		}
		return null;
	    }

	    public int getMetadataIndexByLanguage(String language) {
		for(int i = 0; i < metadatum.size(); i++) {
		    CollectionMeta metadata = (CollectionMeta) metadatum.get(i);
		    if(language.equals(metadata.getLanguage().toString())) {
			return i;
		    }
		}
		return -1;
	    }

	    public Object getValueAt(int row, int column) {
		if(0 <= row && row < metadatum.size()) {
		    CollectionMeta metadata = (CollectionMeta) metadatum.get(row);
		    switch(column) {
		    case 0:
			return metadata.getLanguage();
		    default:
			return metadata.getValue();
		    }
		}
		return "#Error";
	    }

	    public void remove(int index) {
		metadatum.remove(index);
		fireTableRowsDeleted(index, index);
	    }

	    public void setData(ArrayList metadatum) {
		this.metadatum = metadatum;
		fireTableDataChanged();
	    }
	}

	private class FragmentTableRenderer
	    extends DefaultTableCellRenderer {
	    /** Retrieve the component used to rubberstamp the table based on the given parameters.
	     * @param table The <strong>JTable</strong> to rendered.
	     * @param value The <strong>Object</strong> whose representation will be shown in the table row.
	     * @param isSelected <i>true</i> iff the column/row is selected.
	     * @param hasFocus <i>true</i> iff the column/row is in focus.
	     * @param row An <i>int</i> specifying the target row.
	     * @param column An <i>int</i> specifying the target column.
	     * @see org.greenstone.gatherer.cdm.Language
	     */
	    public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected, boolean hasFocus, int row, int column) {
		JComponent component = (JComponent) super.getTableCellRendererComponent(table, value, isSelected, hasFocus, row, column);
		if(isSelected) {
		    component.setBackground(Gatherer.config.getColor("coloring.workspace_heading_background", false));
		}
		else {
		    component.setBackground(Gatherer.config.getColor("coloring.collection_tree_background", false));
		}
		return component;
	    }
	}

	private class LanguageActionListener
	    implements ActionListener {
	    public void actionPerformed(ActionEvent event) {
		if(!ignore_event) {
		    ignore_event = true;
		    ///ystem.err.println("LanguageActionListener");
		    // If the newly selected language value already exists in the fragment table, and that row isn't the one selected, then select that row the text area.
		    String language_name = language_combobox.getSelectedItem().toString();
		    int index = fragment_table_model.getMetadataIndexByLanguage(language_name);
		    if(index != -1) {
			CollectionMeta metadata = fragment_table_model.getMetadata(index);
			fragment_table.setRowSelectionInterval(index, index);
			translation_area.setText(metadata.getValue());
		    }
		    else {
			fragment_table.clearSelection();
			translation_area.setText("");
		    }
		    // Ready the text area
		    translation_area.setEnabled(true);
		    translation_area.setBackground(Gatherer.config.getColor("coloring.editable", false));
		    ignore_event = false;
		}
	    }
	}

	private class TranslationDocumentListener
	    implements DocumentListener {
	    /** Gives notification that an attribute or set of attributes changed. */
	    public void changedUpdate(DocumentEvent e) {
		update();
	    }
	    /** Gives notification that there was an insert into the document. */
	    public void insertUpdate(DocumentEvent e) {
		update();
	    }
	    /** Gives notification that a portion of the document has been removed. */
	    public void removeUpdate(DocumentEvent e) {
		update();
	    }
	    /** The text area has changed in some way. Given that this can only happed when we are editing or adding a text fragment we better respond appropriately. */
	    private void update() {
		if(!ignore_event) {
		    ignore_event = true;
		    String translation_text = translation_area.getText();
		    // If there is no current fragment table selection, but translation text is not an empty string, then this is a new fragment. Create new collection metadata, refresh the fragment table, then ensure that the new metadata is selected. Remember that it is the selected metadata to which changes will be applied.
		    int index = -1;
		    if((index = fragment_table.getSelectedRow()) == -1 && translation_text.length() > 0) {
			///ystem.err.println("Adding new collectionmeta!");
			BobTheMagicalComparableWrapper bob = (BobTheMagicalComparableWrapper)features_list.getSelectedValue();
			Object selected_feature = bob.getContent();
			Language language = (Language) language_combobox.getSelectedItem();
			CollectionMeta metadata = new CollectionMeta(manager, selected_feature, language, translation_text);
			manager.collectionmetadatum.addMetadata(metadata);
			fragment_table.clearSelection();
			fragment_table_model.setData(manager.collectionmetadatum.getMetadata(selected_feature));
			index = fragment_table_model.getMetadataIndexByLanguage(language.toString());
			fragment_table.setRowSelectionInterval(index, index);
			metadata = null;
			language = null;
			selected_feature = null;
		    }
		    else {
			CollectionMeta metadata = fragment_table_model.getMetadata(index);
			// If the text is now empty, and there was previously a metadata fragment selected, then remove the fragments metadata and update the table
			if(translation_text.length() == 0) {
			    ///ystem.err.println("Removing collectionmeta! fireTableRowsDeleted(" + index + "," + index + ")");
			    manager.collectionmetadatum.removeMetadata(metadata);
			    fragment_table_model.remove(index);
			}
			// Otherwise update the fragment metadata and ask the table to repaint the appropriate row
			else {
			    ///ystem.err.println("Updating collectionmeta!");
			    metadata.setValue(translation_text);
			    fragment_table_model.fireTableRowsUpdated(index, index);
			}
			metadata = null;
		    }
		    translation_text = null;
		    ignore_event = false;
		}
	    }
	}
    }

}


