package org.greenstone.gatherer.gui;
/**
 *#########################################################################
 *
 * A component of the Gatherer application, part of the Greenstone digital
 * library suite from the New Zealand Digital Library Project at the
 * University of Waikato, New Zealand.
 *
 * <BR><BR>
 *
 * Author: John Thompson, Greenstone Digital Library, University of Waikato
 *
 * <BR><BR>
 *
 * Copyright (C) 1999 New Zealand Digital Library Project
 *
 * <BR><BR>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * <BR><BR>
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * <BR><BR>
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *########################################################################
 */
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import org.greenstone.gatherer.Gatherer;
import org.greenstone.gatherer.cdm.Argument;
import org.greenstone.gatherer.checklist.CheckList;
import org.greenstone.gatherer.collection.BuildOptions;
import org.greenstone.gatherer.collection.Collection;
import org.greenstone.gatherer.collection.CollectionManager;
import org.greenstone.gatherer.msm.ElementWrapper;
import org.greenstone.gatherer.util.Utility;
import org.w3c.dom.*;
/** This class serves as the data holder for all subclasses of option panes, such as Import options or All options. It also contains methods for creating each of the option lines as they would appear in the subpane. Futhermore it has a method for considering all the arguments and generating a <strong>String[]</strong> to allow you to pass them to the <strong>GShell</strong>.
 * @author John Thompson, Greenstone Digital Library, University of Waikato
 * @version 2.2
 */
final public class OptionsPane
    extends JPanel {
    /** The <strong>BuildOptions</strong> data object contains all the option settings we wish to persist between Gatherer sessions (and thus is stored in <strong>Collection</strong>). */
    private BuildOptions build_options = null;
    /** the log pane - we only create it once now, not each time */
    private JPanel log_pane = null;
    /** the list of previous log messages */
    private JList log_list = null;
    /** whether that currently showing log is changed or not */
    private boolean is_log_changed = false;
    /** the list entry that is currently displayed */
    private FileEntry displayed_log = null;
    /** a listener for changes in the log text */
    private LogChangeListener log_change_listener = null;
    /** All process messages are written to this log text area. */
    private JTextArea log_display = null;

    static final public char SUCCESSFUL = 's';
    static final public char UNSUCCESSFUL = 'u';
    static final public char CANCELLED = 'c';

    static private int BUILD = 0;
    static private int IMPORT = 1;
    static private Dimension LABEL_SIZE = new Dimension(180, 25);
    static private Dimension LOG_SIZE = new Dimension(610, 360);
    /** The minimum size of a options pane, to prevent refresh problems around the bottom of shorter panes. */
    static private Dimension PANE_SIZE = new Dimension(610, 350);
    static private Dimension ROW_SIZE = new Dimension(610, 25);
    static private Dimension SPINNER_SIZE = new Dimension(100, 25);
    static private String DESCRIPTION_SEP = " + ";


	 /** The default constructor creates the few session length options, but either retrieves the rest from the current collection, or creates a default set of options. */
    public OptionsPane(JTextArea log, BuildOptions build_options) {
	this.log_display = log;
	this.build_options = build_options;
	this.log_change_listener = new LogChangeListener();
	//this.target_pane = target_pane;
    }

    /** creates a new log from the text in log. The new file is named
     * build_log.<current time>.txt. A new FileEntry is added to teh log list
     * and made selected. A fresh listener is added to teh document */
    public void addNewLog(char success) {
	///ystem.out.println("add new log");

	// create the file name
	long time = System.currentTimeMillis();
	StringBuffer name = new StringBuffer();
	name.append("build_log.");
	name.append(time);
	name.append(success);
	name.append(".txt");
	File f = new File(Gatherer.c_man.getCollectionLog()+name.toString());
	// just in case there is no log directory
	File parent_file = f.getParentFile();
	parent_file.mkdirs();

	// save the log text to the file
	saveLogFile(f);
	is_log_changed=false;
	log_display.getDocument().addDocumentListener(log_change_listener);

	// create the file entry and add it to the list at pos 0 - it will always be the newest one created
	FileEntry fe = new FileEntry(f);
	((DefaultListModel)log_list.getModel()).add(0, fe);
	log_list.setSelectedIndex(0);
	displayed_log = fe;

    }

    /** This method creates the panel with all the build only options on it.
	  * @return A <strong>JPanel</strong> which can be used to display all the build only options.
	  */
    public JPanel buildBuild() {
	JPanel pane = new JPanel();
	pane.setBackground(Gatherer.config.getColor("coloring.collection_tree_background", false));
	pane.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));
	int build_argument_count = build_options.getBuildArgumentCount();
	pane.setLayout(new GridLayout(build_argument_count, 1, 0, 0));
	int total_height = PANE_SIZE.height;
	for(int i = 0; i < build_argument_count; i++) {
				// Retrieve the argument so we know how to format the control.
	    Argument argument = build_options.getBuildArgument(i);
				// Now attempt to retrieve any existing value for this argument.
	    boolean enabled = build_options.getBuildValueEnabled(argument.getName());
	    String value = build_options.getBuildValue(argument.getName());
	    ArgumentControl argument_control = new ArgumentControl(BUILD, argument, enabled, value);
	    total_height = total_height - argument_control.getPreferredSize().height;
	    pane.add(argument_control);
	}
	if(total_height > 0) {
	    JPanel filler = new JPanel();
	    filler.setPreferredSize(new Dimension(100, total_height));
	    filler.setSize(filler.getPreferredSize());
	    pane.add(filler);
	}
	return pane;
    }
    /** This method creates the panel with all the import only options on it.
	  * @return A <strong>JPanel</strong> which can be used to display all the import only options.
	  */
    public JPanel buildImport() {
	JPanel pane = new JPanel();
	pane.setBackground(Gatherer.config.getColor("coloring.collection_tree_background", false));
	pane.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));
	int import_argument_count = build_options.getImportArgumentCount();
	pane.setLayout(new GridLayout(import_argument_count, 1, 0, 0));
	int total_height = PANE_SIZE.height;
	for(int i = 0; i < import_argument_count; i++) {
				// Retrieve the argument so we know how to format the control.
	    Argument argument = build_options.getImportArgument(i);
				// Now attempt to retrieve any existing value for this argument.
	    boolean enabled = build_options.getImportValueEnabled(argument.getName());
	    String value = build_options.getImportValue(argument.getName());
	    ArgumentControl argument_control = new ArgumentControl(IMPORT, argument, enabled, value);
	    total_height = total_height - argument_control.getPreferredSize().height;
	    pane.add(argument_control);
	}
	if(total_height > 0) {
	    JPanel filler = new JPanel();
	    filler.setPreferredSize(new Dimension(100, total_height));
	    filler.setSize(filler.getPreferredSize());
	    pane.add(filler);
	}
	return pane;
    }
    /** This method is used to build a panel based on the message log, which is nothing like any of the other panels.
	  * @return A <strong>JPanel</strong> containing a scrollable text area which represents the shell process message log.
	  */
    public JPanel buildLog() {
	// we now save the log pane
	if (log_pane == null) {
	    log_pane = new JPanel(new BorderLayout());

	    // Build a list of the log files available, ordering by last modified. Log files are like build_log.date.txt
	    DefaultListModel contents = new DefaultListModel();
	    File log_directory = new File(Gatherer.c_man.getCollectionLog());
	    File children[] = log_directory.listFiles();
	    for(int i = 0; children != null && i < children.length; i++) {
		if(children[i].getName().startsWith("build_log")&&children[i].getName().endsWith(".txt") ) {
		    FileEntry entry = new FileEntry(children[i]);
		    // We are about to insert it. But where.
		    boolean found = false;
		    for(int j = 0; j < contents.size(); j++) {
			FileEntry sibling = (FileEntry) contents.getElementAt(j);
			int order = entry.compareTo(sibling);
			if(order > 0) {
			    contents.insertElementAt(entry, j);
			    found = true;
			    break;
			}
		    }
		    if(!found) {
			contents.addElement(entry);
		    }
		}
	    }

	    log_list = new JList(contents);
	    log_list.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
	    log_list.setLayoutOrientation(JList.VERTICAL);
	    log_list.setVisibleRowCount(3);
	    log_list.addListSelectionListener(new LogListListener());

	    JScrollPane list_scroller = new JScrollPane(log_list);

	    // If log is not empty and we have something in displayed_log - make that element selected
	    if (log_display.getText().length()>0 && displayed_log != null) {
		int index = contents.indexOf(displayed_log);
		if (index != -1) {
		    log_list.setSelectedIndex(index);
		}

	    }
	    log_pane.add(new JScrollPane(log_display), BorderLayout.CENTER);
	    JLabel log_history_label = new JLabel(get("LogHistory"));
	    JPanel log_history_pane = new JPanel();
	    log_history_pane.setLayout(new BorderLayout());
	    log_history_pane.add(log_history_label, BorderLayout.NORTH);
	    log_history_pane.add(list_scroller, BorderLayout.SOUTH);
	    log_pane.add(log_history_pane, BorderLayout.SOUTH);
	}
	return log_pane;
    }

    /** This method retrieves a phrase from the dictionary based apon the key.
	  * @param key A <strong>String</strong> used as a reference to the correct phrase.
	  * @return A phrase, matching the key, as a <strong>String</strong>.
	  */
    private String get(String key) {
	return get(key, null);
    }
    /** This method retrieves a phrase from the dictionary based apon the key and arguments.
	  * @param key A <strong>String</strong> used as a reference to the correct phrase.
	  * @param args A <strong>String[]</strong> whose contents are often substituted into the phrase before it is returned.
	  * @return A phrase, matching the key and constructed using the arguments, as a <strong>String</strong>.
	  * @see org.greenstone.gatherer.Dictionary
	  * @see org.greenstone.gatherer.Gatherer
	  */
    private String get(String key, String args[]) {
	if(key.indexOf(".") == -1) {
	    key = "OptionsPane." + key;
	}
	return Gatherer.dictionary.get(key, args);
    }
    /** Attempts to discover the latest document count.
	  * @return An <strong>int</strong> detailing the number of documents in this collection.
	 */
    public int getDocumentCount() {
	if(Gatherer.c_man.ready()) {
	    int count = Gatherer.c_man.getCollection().getDocumentCount();
	    if(count != 0) {
		return count;
	    }
	}
	return 1;
    }

    /** Loads a new log into the log text - saves the currently showing one,
     * and loads the new one, adding a fresh listener to the document
     * Note, we need to remove any listener before making the changes, then
     * add a listener back in
     */
    public void loadSelectedLog(FileEntry fe) {
	///ystem.out.println("load selected log");
	if (is_log_changed) {
	    saveLogFile(displayed_log.getFile());
	} else {
	    log_display.getDocument().removeDocumentListener(log_change_listener);
	}

	displayed_log = fe;
	if (fe != null) {
	    log_display.setText(fe.getFileText());
	} else {
	    log_display.setText("");
	}
	is_log_changed=false;
	log_display.getDocument().addDocumentListener(log_change_listener);

    }

    /** Neutralizes the log list - saves any unsaved changes to the current log
     * clears the log text, unselects the log list, and adds a fresh listener
     * to the log's document
     * Note, we need to remove any listener before making the changes, then
     * add a listener back in */
    public void resetLogList() {
	///ystem.out.println("reset log");
	if (is_log_changed) {
	    saveLogFile(displayed_log.getFile());
	} else {
	    log_display.getDocument().removeDocumentListener(log_change_listener);
	}
	log_display.setText("");
	displayed_log = null;
	if (log_list != null) {
	    log_list.clearSelection();
	}
	is_log_changed=false;
	log_display.getDocument().addDocumentListener(log_change_listener);

    }

    /** saves the current text in log to the File specified */
    public void saveLogFile(File file) {

	try {
	    BufferedWriter out = new BufferedWriter(new FileWriter(file)); // uses the default encoding
	    String log_text = log_display.getText();
	    // set teh return delims to true, then swap them for newLine() (platform dependent) - having return delims set to false seemed to not get empty strings so that empty lines were not preserved.
	    StringTokenizer t = new StringTokenizer(log_text, "\n", true);
	    while (t.hasMoreTokens()) {
		String token = t.nextToken();
		if (token.equals("\n")) {
		    out.newLine();
		} else {
		    out.write(token, 0, token.length());
		}

	    }
	    out.close();

	} catch(Exception e) {
	    System.out.println("Error in saving log file: "+file.toString());
	    e.printStackTrace();
	}


    }


    /** Given a panel containing ArgumentControls, update the values associated with them. */
    public void update(JPanel panel) {
	if(panel == log_pane) {
	    return;
	}

	for(int i = 0; i < panel.getComponentCount(); i++) {
	    Component component = panel.getComponent(i);
	    if(component instanceof ArgumentControl) {
		((ArgumentControl)component).update();
	    }
	}
    }

    private class ArgumentControl
	extends JPanel {
	private Argument argument;
	private int type;
	private JComponent value_control;
	private JCheckBox enabled;
	public ArgumentControl(int type, Argument argument, boolean enable, String value) {
	    super();
	    this.argument = argument;
	    this.type = type;
	    String tooltip = Utility.formatHTMLWidth("<html>" + argument.getDesc() + "</html>", 60);
				// Because of the dynamic order of component creation/connection/layout, we can't really follow that pattern here.
	    setBackground(Gatherer.config.getColor("coloring.collection_tree_background", false));
	    setBorder(BorderFactory.createEmptyBorder(2,0,2,0));
	    setLayout(new BorderLayout());
	    setPreferredSize(ROW_SIZE);

				/*
				  JPanel inner_pane = new JPanel();
				  inner_pane.setOpaque(false);
				*/

				// Try to determine what the value_controls value should be.
	    if(value == null) {
		value = argument.getDefaultValue();
	    }
				// Create a correct value control based on the argument provided.
	    switch(argument.getType()) {
	    case Argument.ENUM:
		JComboBox combobox = new JComboBox();
		combobox.setEnabled(enable);
		combobox.setToolTipText(tooltip);
		// Set enabled
		if(enable) {
		    combobox.setBackground(Color.white);
		}
		else {
		    combobox.setBackground(Color.lightGray);
		}
		// Build an option model, wrapping each entry of the list table.
		HashMap arg_list = argument.getList();
		Iterator it = arg_list.keySet().iterator();
		while(it.hasNext()) {
		    combobox.addItem((String) it.next());
		}
		// Connect this up first, so that if a value is selected the tooltip updates accordingly.
		combobox.addActionListener(new ToolTipUpdater(arg_list));
		if(value != null) {
		    // Set the selected string. However since they are all strings we had best iterate ourselves.
		    for(int i = 0; i < combobox.getItemCount(); i++) {
			if(combobox.getItemAt(i).toString().equals(value)) {
			    combobox.setSelectedIndex(i);
			}
		    }
		}
		// Layout
		add(combobox, BorderLayout.CENTER);
		// And remember
		value_control = combobox;
		break;
	    case Argument.FLAG:
		// Only need the check box.
		value_control = null;
		break;
	    case Argument.INTEGER:
		// Build a spinner
		JSpinner spinner = new JSpinner();
		spinner.setEnabled(enable);
		spinner.setPreferredSize(SPINNER_SIZE);
		spinner.setToolTipText(tooltip);
		// Set enabled
		JComponent c = spinner.getEditor();
		if ( c instanceof JSpinner.DefaultEditor ) {
		    JSpinner.DefaultEditor editor = (JSpinner.DefaultEditor) c;
		    JFormattedTextField field = editor.getTextField();
		    field.setEditable(enable);
		    if(enable) {
			field.setBackground(Color.white);
		    }
		    else {
			field.setBackground(Color.lightGray);
		    }
		}
		// If there was an original value, set it.
		if(value != null) {
		    try {
			spinner.setValue(new Integer(value));
		    }
		    catch (Exception error) {
		    }
		}
		// Layout
		add(new JLabel(), BorderLayout.CENTER);
		add(spinner, BorderLayout.EAST);
		// And remember it
		value_control = spinner;
		break;
	    case Argument.STRING:
		// Use a standard text field
		JTextField textfield = new JTextField();
		textfield.setEnabled(enable);
		textfield.setToolTipText(tooltip);
		// Set enabled
		if(enable) {
		    textfield.setBackground(Color.white);
		}
		else {
		    textfield.setBackground(Color.lightGray);
		}
		// If there was an original value, set it.
		if(value != null) {
		    textfield.setText(value);
		}
		// Layout
		add(textfield, BorderLayout.CENTER);
		// And remember it
		value_control = textfield;
		break;
	    }

				// If the argument is required, then you don't get a choice of whether it is enabled.
	    if(argument.isRequired()) {
		JLabel label = new JLabel(argument.getName());
		label.setOpaque(false);
		label.setPreferredSize(LABEL_SIZE);
		label.setToolTipText(tooltip);
		add(label, BorderLayout.WEST);
	    }
	    else {
		enabled = new JCheckBox(argument.getName(), enable);
		enabled.setOpaque(false);
		enabled.setPreferredSize(LABEL_SIZE);
		enabled.setToolTipText(tooltip);
		// Connect
		enabled.addActionListener(new EnabledListener(value_control));
		// Layout
		add(enabled, BorderLayout.WEST);
	    }
	}

	/** Update the values stored in the collection so as to rememebr the current state of this argument. */
	public void update() {
	    String name = argument.getName();
	    boolean enable = true;
	    if(enabled != null) {
		enable = enabled.isSelected();
	    }
	    String value = null;
	    if(value_control == null) {
		// Flag value, nothing to do.
	    }
	    else if(value_control instanceof JTextField) {
		value = ((JTextField)value_control).getText();
	    }
	    else if(value_control instanceof JSpinner) {
		value = ((JSpinner)value_control).getValue().toString();
	    }
	    else if(value_control instanceof JComboBox) {
		value = (String) ((JComboBox)value_control).getSelectedItem();
	    }
				// If this argument was a flag, but is now disabled, remove from the build options altogether
	    if(!enable && value == null) {
		if(type == BUILD) {
		    build_options.removeBuildValue(name);
		}
		else {
		    build_options.removeImportValue(name);
		}
	    }
				// Otherwise update the argument value
	    else {
		if(type == BUILD) {
		    build_options.setBuildValue(name, enable, value);
		}
		else {
		    build_options.setImportValue(name, enable, value);
		}
	    }
	}
    }

    /** Listens for actions apon the enable checkbox, and if detected enables or diables control appropriately. */
    private class EnabledListener
	implements ActionListener {
	/** An editor component, such as a JComboBox or JTextField, that might have its enabled state changed by this listener. */
	private JComponent target = null;
	/** Constructor. */
	public EnabledListener(JComponent target) {
	    this.target = target;
	}
	/** Any implementation of ActionListener must include this method so that we can be informed when an action has been performed on or registered check box, prompting us to change the state of the other controls as per the users request.
			* @param event An <strong>ActionEvent</strong> containing information about the click.
			*/
	public void actionPerformed(ActionEvent event) {
	    JCheckBox source = (JCheckBox)event.getSource();
	    if(target != null) {
		if(source.isSelected()) {
		    target.setBackground(Color.white);
		    target.setEnabled(true);
		}
		else {
		    target.setBackground(Color.lightGray);
		    target.setEnabled(false);
		}
		// Special case of stupid JSpinners who don't let their backgrounds change properly.
		if(target instanceof JSpinner) {
		    JSpinner spinner = (JSpinner) target;
		    JComponent c = spinner.getEditor();
		    if ( c instanceof JSpinner.DefaultEditor ) {
			JSpinner.DefaultEditor editor = (JSpinner.DefaultEditor) c;
			JFormattedTextField field = editor.getTextField();
			field.setEditable(source.isSelected());
			if(source.isSelected()) {
			    field.setBackground(Color.white);
			}
			else {
			    field.setBackground(Color.lightGray);
			}
		    }
		}
	    }
	}
    }

    /** holds a File which has a particular naming convention
     * build_log.date.txt
     * also keeps a Date corresponding to the date in its name*/
    private class FileEntry {

	private Date date=null;
	private File file=null;
	private String display=null;
	public FileEntry(File f) {
	    this.file=f;
	    this.date = getDateFromFileName();
	    this.display = createDisplayText();

	}

	/** returns 0 if the dates are the same, -ve number if the current FileEntry is earlier than the fe FileEntry ...*/
	public int compareTo(FileEntry fe) {
	    Date d = fe.getDate();
	    return date.compareTo(d);
	}

	public Date getDate() {
	    return this.date;
	}
	public File getFile() {
	    return this.file;
	}
	/** returns the contents of the file */
	public String getFileText() {
	    if (!this.file.exists()) {
		return "";
	    }
	    StringBuffer contents = new StringBuffer();
	    try {
		BufferedReader in = new BufferedReader(new FileReader(this.file));
		String line;

		while ((line = in.readLine()) != null) {
		    contents.append(line);
		    contents.append("\n");
		}
	    } catch (FileNotFoundException fnfe) {
		Gatherer.println("Error: log file not found: "+ this.file.toString());
		return "";
	    } catch (IOException ioe) {
		Gatherer.println("Error: exception occurred when trying to read in file "+this.file.toString()+", "+ioe.getMessage());

	    }

	    return contents.toString();
	}
	/** we only want the date out of the file name, not the whole path */
	public String toString() {
	    return this.display;
	}
	/** creates the display text for the list */
	private String createDisplayText() {
	    StringBuffer d = new StringBuffer();
	    d.append(date.toString());
	    //char success= this.file.getName().charAt(23);
	    String filename = this.file.getName();
	    int index = filename.lastIndexOf(".") - 1;
	    if(index >= 0) {
	    	char success = filename.charAt(index);
	    	switch (success) {
	    	case OptionsPane.SUCCESSFUL:
			d.append(get("Successful"));
			break;
	    	case OptionsPane.UNSUCCESSFUL:
			d.append(get("Unsuccessful"));
			break;
	    	case OptionsPane.CANCELLED:
			d.append(get("Cancelled"));
			break;
	    	}
		}
	    return d.toString();
	}
	/** extracts the date from the file name, which is of the form 'build_log.'<long>[scu]'.txt' */
	private Date getDateFromFileName() {
	    String name = this.file.getName();
	    // Need to exclude first '.'
	    int first_index = name.indexOf(".") + 1;
	    // Need to exclude state letter before last '.'
	    int last_index = name.lastIndexOf(".") - 1;
	    if(first_index > 0 && last_index > 0 && first_index < last_index) {
		String date_string = name.substring(first_index, last_index);
		return new Date(Long.parseLong(date_string));
	    }
	    else {
		return new Date(); // Current date
	    }
	}


    }

    /** A DocumentListener that listens for changes in the log_display document
     * if a change is registered, the flag is_log_changed is set to true,
     * and the listener removes itself from log_display
     * Because we only want to register user's changes, the listener must be removed before the system makes any changes, and then put back. For example, each time we switchlogs to display, a remove event and an insert event are sent. We dont want to register these as changes to the document, so the listener must be removed before the switch, and added back afterwards */
    private class LogChangeListener implements DocumentListener {

	public void changedUpdate(DocumentEvent e) {
	    ///ystem.out.println("changed update");
	    registerChange();
	}
	public void insertUpdate(DocumentEvent e) {
	    ///ystem.out.println("insert update");
	    registerChange();
	}
	public void removeUpdate(DocumentEvent e) {
	    ///ystem.out.println("remove update");
	    registerChange();
	}

	private void registerChange() {
	    ///ystem.err.println("a change in the doc occurred, removing the listener");
	    is_log_changed = true;
	    log_display.getDocument().removeDocumentListener(this);
	}
    }

    /** a ListSelectionListener that triggers the load of a newly selected
	log */
    private class LogListListener implements ListSelectionListener {

	public void valueChanged(ListSelectionEvent e) {
	    if (!e.getValueIsAdjusting()) { // we get two events for one change in list selection - use the false one ( the second one)
		JList source  = (JList)e.getSource();
		FileEntry fe = (FileEntry) source.getSelectedValue();
		loadSelectedLog(fe);
	    }
	}
    }

    /** Listener that sets the tooltip associated to a combobox to the tooltip relevant to the selected item. */
    private class ToolTipUpdater
	implements ActionListener {
	private HashMap arg_list;
	public ToolTipUpdater(HashMap arg_list) {
	    this.arg_list = arg_list;
	}
	/** Any implementation of an ActionListener must include this method so that we can be informed when the selection in a combobox has changed and update the tooltip accordingly.
	 * @param event An <strong>ActionEvent</strong> containing information about the action that fired this call.
	 */
	public void actionPerformed(ActionEvent event) {
	    JComboBox source = (JComboBox)event.getSource();
	    String key = (String) source.getSelectedItem();
	    if(arg_list != null) {
		String description = (String) arg_list.get(key);
		if(description != null) {
		    description = Utility.formatHTMLWidth(DESCRIPTION_SEP + description, 60);
		    String original = source.getToolTipText();
		    if(original == null) {
			original = "";
		    }
		    // Remove any existing extra description.
		    if(original.indexOf(DESCRIPTION_SEP) != -1) {
			original = original.substring(0, original.indexOf(DESCRIPTION_SEP));
		    }
		    source.setToolTipText(original + description);
		}
	    }
	}
    }
}
